"""
methods to interact with farcaster hub
"""

# lib
import os
import requests

# src
from .models import FrameMessage, ValidatedMessage, ValidatedData, FrameAction, CastId


def get_message(
        msg: str,
        hub: str,
        username: str = None,
        password: str = None,
        api_key: str = None
) -> ValidatedMessage:
    url = f'{hub}/v1/validateMessage'
    headers = {'content-type': 'application/octet-stream'}
    if api_key:
        headers['api_key'] = api_key
    auth = None
    if username:
        auth = (username, password)
    res = requests.post(url, headers=headers, auth=auth, data=bytes.fromhex(msg))

    body = res.json()
    if not body['valid']:
        raise ValueError('frame action message is invalid')
    action = ValidatedMessage(**body['message'])

    return action


def validate_message(
        msg: FrameMessage,
        hub: str,
        username: str = None,
        password: str = None,
        api_key: str = None
) -> ValidatedMessage:
    action = get_message(msg.trustedData.messageBytes, hub, username=username, password=password, api_key=api_key)

    if msg.untrustedData.fid != action.data.fid:
        raise ValueError(f'fid does not match: {msg.untrustedData.fid} {action.data.fid}')

    if msg.untrustedData.buttonIndex != action.data.frameActionBody.buttonIndex:
        raise ValueError(
            f'button index does not match: {msg.untrustedData.buttonIndex} {action.data.frameActionBody.buttonIndex}')

    if msg.untrustedData.inputText is not None and msg.untrustedData.inputText != action.data.frameActionBody.inputText:
        raise ValueError(
            f'text input does not match: {msg.untrustedData.inputText} {action.data.frameActionBody.inputText}')

    if msg.untrustedData.state is not None and msg.untrustedData.state != action.data.frameActionBody.state:
        raise ValueError(f'state does not match: {msg.untrustedData.state} {action.data.frameActionBody.state}')

    return action


def validate_message_or_mock(
        msg: FrameMessage,
        hub: str,
        username: str = None,
        password: str = None,
        api_key: str = None,
        mock: bool = False
) -> ValidatedMessage:
    if mock:
        # mock
        return ValidatedMessage(
            data=ValidatedData(
                type='MESSAGE_TYPE_FRAME_ACTION',
                fid=msg.untrustedData.fid,
                timestamp=msg.untrustedData.timestamp,
                network=str(msg.untrustedData.network),
                frameActionBody=FrameAction(
                    url=msg.untrustedData.url,
                    buttonIndex=msg.untrustedData.buttonIndex,
                    castId=msg.untrustedData.castId,
                    inputTest=msg.untrustedData.inputText,
                    state=msg.untrustedData.state,
                    transactionId=msg.untrustedData.transactionId
                )
            ),
            hash=msg.untrustedData.messageHash,
            hashScheme='HASH_SCHEME_BLAKE',
            signature='',
            signature_scheme='SIGNATURE_SCHEME_ED25519',
            signer=''
        )

    return validate_message(msg, hub, username=username, password=password, api_key=api_key)



"""
framelib module
"""

from .frame import frame, message
from .hub import validate_message, validate_message_or_mock
from .models import FrameMessage, ValidatedMessage, User
from .warpcast import get_user
from .neynar import (
    validate_message as validate_message_neynar,
    validate_message_or_mock as validate_message_or_mock_neynar
)
from .transaction import transaction



"""
methods to call neynar api
"""

# lib
import os
import requests

# src
from .models import FrameMessage, NeynarValidatedMessage, NeynarInteractor, NeynarProfile, NeynarBio, \
    NeynarButton, NeynarInput, NeynarState, NeynarTransaction


def get_frame_message(msg: str, api_key: str) -> NeynarValidatedMessage:
    if not api_key:
        raise ValueError('neynar api key not set')
    url = 'https://api.neynar.com/v2/farcaster/frame/validate'
    body = {
        'cast_reaction_context': False,  # TODO
        'follow_context': False,
        'signer_context': False,
        'message_bytes_in_hex': msg
    }
    headers = {
        'accept': 'application/json',
        'api_key': api_key,
        'content-type': 'application/json'
    }
    res = requests.post(url, json=body, headers=headers)

    body = res.json()
    if not body['valid']:
        raise ValueError('frame action message is invalid')

    action = NeynarValidatedMessage(**body['action'])

    return action


def validate_message(msg: FrameMessage, api_key: str) -> NeynarValidatedMessage:
    action = get_frame_message(msg.trustedData.messageBytes, api_key)

    if msg.untrustedData.fid != action.interactor.fid:
        raise ValueError(f'fid does not match: {msg.untrustedData.fid} {action.interactor.fid}')

    if msg.untrustedData.buttonIndex != action.tapped_button.index:
        raise ValueError(f'button index does not match: {msg.untrustedData.buttonIndex} {action.tapped_button.index}')

    if msg.untrustedData.inputText is not None and msg.untrustedData.inputText != action.input.text:
        raise ValueError(f'text input does not match: {msg.untrustedData.inputText} {action.input.text}')

    if msg.untrustedData.state is not None and msg.untrustedData.state != action.state.serialized:
        raise ValueError(f'state does not match: {msg.untrustedData.state} {action.state.serialized}')

    return action


def validate_message_or_mock(msg: FrameMessage, api_key: str, mock: bool = False) -> NeynarValidatedMessage:
    if mock:
        # mock
        # TODO option to populate with warpcast profile
        return NeynarValidatedMessage(
            object='validated_frame_action',
            interactor=NeynarInteractor(
                object='user',
                fid=msg.untrustedData.fid,
                username=f'username {msg.untrustedData.fid}',
                display_name=f'display name {msg.untrustedData.fid}',
                pfp_url='',
                profile=NeynarProfile(bio=NeynarBio(text='')),
                follower_count=0,
                following_count=0,
                verifications=['0x'],
                active_status='',
            ),
            tapped_button=NeynarButton(index=msg.untrustedData.buttonIndex),
            input=NeynarInput(text=msg.untrustedData.inputText or ''),  # TODO set model to None if missing
            state=NeynarState(serialized=msg.untrustedData.state or ''),
            transaction=NeynarTransaction(hash=msg.untrustedData.transactionId or ''),
            url=msg.untrustedData.url,
            timestamp=msg.untrustedData.timestamp,
            cast={}
        )

    return validate_message(msg, api_key)



"""
methods for frame rendering and message handling
"""

# lib
import json
from importlib import resources
from typing import Literal
from flask import render_template_string, request, make_response, Response

# src
from .models import FrameMessage

# enum types
ButtonActions = Literal['post', 'post_redirect', 'mint', 'link', 'tx']
AspectRatio = Literal['1.91:1', '1:1']


def frame(
        title: str = None,
        image: str = None,
        aspect_ratio: AspectRatio = None,
        content: str = None,
        post_url: str = None,
        button1: str = None,
        button1_action: ButtonActions = None,
        button1_target: str = None,
        button2: str = None,
        button2_action: ButtonActions = None,
        button2_target: str = None,
        button3: str = None,
        button3_action: ButtonActions = None,
        button3_target: str = None,
        button4: str = None,
        button4_action: ButtonActions = None,
        button4_target: str = None,
        input_text: str = None,
        state: str = None,
        max_age: int = None
) -> Response:
    # setup context
    # note: we do this because jinja treats None as a defined value
    ctx = {k: v for k, v in locals().items() if v is not None}

    # load template from module data
    pth = resources.files('framelib') / 'templates' / 'frame.html'
    with open(str(pth), 'r') as f:
        src = f.read()

    # render frame template
    html = render_template_string(src, **ctx)

    # response
    res = make_response(html)
    res.status_code = 200
    if max_age is not None:
        res.cache_control.max_age = max_age
    return res


def message() -> FrameMessage:
    # parse action message
    body = json.loads(request.data)
    msg = FrameMessage(**body)
    return msg



"""
methods for frame transactions
"""

# lib
from eth_abi import encode
from eth_utils import is_address, function_signature_to_4byte_selector, function_abi_to_4byte_selector
from flask import jsonify, Response

# src
from .models import Transaction, EthTransactionParams


def transaction(
        chain_id: int,
        contract: str,
        abi: list[dict],
        value: str = None,
        function_signature: str = None,
        function_arguments: list = None
) -> Response:
    if not is_address(contract):
        raise ValueError(f'invalid contract address {contract}')

    # encode transaction calldata
    data = None
    if function_signature:
        fx_abi = None
        for a in abi:
            if 'name' not in a:
                continue
            if function_signature_to_4byte_selector(function_signature) == function_abi_to_4byte_selector(a):
                fx_abi = a
                break
        if fx_abi is None:
            raise ValueError(f'method {function_signature} not found in abi')

        data = '0x' + function_abi_to_4byte_selector(fx_abi).hex()

        if fx_abi['inputs'] and function_arguments:
            data += encode([i['type'] for i in fx_abi['inputs']], function_arguments).hex()

    # setup frame transaction
    tx = Transaction(
        chainId=f'eip155:{chain_id}',
        method='eth_sendTransaction',
        params=EthTransactionParams(abi=abi, to=contract, value=value, data=data)
    )

    # response
    res = jsonify(tx.model_dump(mode='json', exclude_none=True))
    res.status_code = 200
    return res



"""
methods to query farcaster data from the warpcast api
"""

import requests

from .models import User


def get_user(fid: int) -> User:
    res = requests.get('https://client.warpcast.com/v2/user', params={'fid': fid})
    if res.status_code != 200:
        raise ValueError(f'could not get warpcast user profile for fid {fid}')
    return User(**res.json()['result']['user'])



"""
data models
"""

import datetime
from typing import Optional, Literal
from pydantic import BaseModel


# ---- frame message ----

class CastId(BaseModel):
    fid: int
    hash: str


class FrameAction(BaseModel):
    url: str
    buttonIndex: int
    inputText: Optional[str] = None
    state: Optional[str] = None
    transactionId: Optional[str] = None
    castId: CastId


class UntrustedData(FrameAction):
    # note: this untrusted message seems to collapse the ValidatedDate and FrameAction fields
    fid: int
    messageHash: str
    timestamp: int
    network: int


class TrustedData(BaseModel):
    messageBytes: str


class FrameMessage(BaseModel):
    untrustedData: UntrustedData
    trustedData: TrustedData


# ---- frame transaction ----

class EthTransactionParams(BaseModel):
    abi: list[dict]
    to: str
    value: Optional[str]
    data: Optional[str]


class Transaction(BaseModel):
    chainId: str
    method: Literal['eth_sendTransaction']
    params: EthTransactionParams


# ---- hub ----

class ValidatedData(BaseModel):
    type: str
    fid: int
    timestamp: datetime.datetime
    network: str
    frameActionBody: FrameAction


class ValidatedMessage(BaseModel):
    data: ValidatedData
    hash: str
    hashScheme: str
    signature: str
    signatureScheme: str
    signer: str


# ---- neynar ----

class NeynarViewer(BaseModel):
    following: bool
    followed_by: bool


class NeynarBio(BaseModel):
    text: str
    mentioned_profiles: Optional[list[str]] = []


class NeynarProfile(BaseModel):
    bio: NeynarBio


class NeynarInteractor(BaseModel):
    object: str
    fid: int
    username: str
    display_name: str
    custody_address: Optional[str] = None
    pfp_url: str
    profile: NeynarProfile
    follower_count: int
    following_count: int
    verifications: list[str]
    active_status: str
    viewer_context: Optional[NeynarViewer] = None


class NeynarButton(BaseModel):
    title: Optional[str] = None
    index: int
    action_type: Optional[str] = None


class NeynarInput(BaseModel):
    text: str


class NeynarState(BaseModel):
    serialized: str


class NeynarTransaction(BaseModel):
    hash: str


class NeynarValidatedMessage(BaseModel):
    object: str
    interactor: NeynarInteractor
    tapped_button: NeynarButton
    input: Optional[NeynarInput] = None
    state: Optional[NeynarState] = None
    url: str
    cast: dict
    timestamp: datetime.datetime
    transaction: Optional[NeynarTransaction] = None


# ---- warpcast ----

class Pfp(BaseModel):
    url: str
    verified: bool


class WarpBio(BaseModel):
    text: str
    mentions: Optional[list[str]] = []
    channelMentions: Optional[list[str]] = []


class WarpLocation(BaseModel):
    placeId: str
    description: str


class WarpProfile(BaseModel):
    bio: WarpBio
    location: WarpLocation


class User(BaseModel):
    fid: int
    username: Optional[str] = None
    displayName: str
    pfp: Optional[Pfp] = None
    profile: WarpProfile
    followerCount: int
    followingCount: int
    activeOnFcNetwork: bool



"""
test cases for frame transaction logic
"""

# lib
import json
import pytest
from flask import Flask

# src
from framelib import transaction

# constants
ABI_WETH = '[{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"src","type":"address"},{"indexed":true,"internalType":"address","name":"guy","type":"address"},{"indexed":false,"internalType":"uint256","name":"wad","type":"uint256"}],"name":"Approval","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"dst","type":"address"},{"indexed":false,"internalType":"uint256","name":"wad","type":"uint256"}],"name":"Deposit","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"src","type":"address"},{"indexed":true,"internalType":"address","name":"dst","type":"address"},{"indexed":false,"internalType":"uint256","name":"wad","type":"uint256"}],"name":"Transfer","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"src","type":"address"},{"indexed":false,"internalType":"uint256","name":"wad","type":"uint256"}],"name":"Withdrawal","type":"event"},{"payable":true,"stateMutability":"payable","type":"fallback"},{"constant":true,"inputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"address","name":"","type":"address"}],"name":"allowance","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"guy","type":"address"},{"internalType":"uint256","name":"wad","type":"uint256"}],"name":"approve","outputs":[{"internalType":"bool","name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[{"internalType":"address","name":"","type":"address"}],"name":"balanceOf","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"decimals","outputs":[{"internalType":"uint8","name":"","type":"uint8"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[],"name":"deposit","outputs":[],"payable":true,"stateMutability":"payable","type":"function"},{"constant":true,"inputs":[],"name":"name","outputs":[{"internalType":"string","name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"symbol","outputs":[{"internalType":"string","name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"totalSupply","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"dst","type":"address"},{"internalType":"uint256","name":"wad","type":"uint256"}],"name":"transfer","outputs":[{"internalType":"bool","name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"src","type":"address"},{"internalType":"address","name":"dst","type":"address"},{"internalType":"uint256","name":"wad","type":"uint256"}],"name":"transferFrom","outputs":[{"internalType":"bool","name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"internalType":"uint256","name":"wad","type":"uint256"}],"name":"withdraw","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"}]'

app = Flask(__name__)


class TestTransaction(object):

    def test_deposit(self):
        abi = json.loads(ABI_WETH)
        with app.app_context():
            res = transaction(8453, '0x4200000000000000000000000000000000000006', abi,
                              value='50000000000000000', function_signature='deposit()')

            assert res.status_code == 200
            assert res.json['chainId'] == 'eip155:8453'
            assert res.json['method'] == 'eth_sendTransaction'
            assert res.json['params']['abi'] == abi
            assert res.json['params']['to'] == '0x4200000000000000000000000000000000000006'
            assert res.json['params']['value'] == '50000000000000000'
            assert res.json['params']['data'] == '0xd0e30db0'  # function selector

    def test_withdraw(self):
        abi = json.loads(ABI_WETH)
        with app.app_context():
            res = transaction(8453, '0x4200000000000000000000000000000000000006', abi,
                              function_signature='withdraw(uint256)', function_arguments=[int(123e18)])

            assert res.status_code == 200
            assert res.json['chainId'] == 'eip155:8453'
            assert res.json['method'] == 'eth_sendTransaction'
            assert res.json['params']['abi'] == abi
            assert res.json['params']['to'] == '0x4200000000000000000000000000000000000006'
            assert 'value' not in res.json['params']
            assert res.json['params']['data'] \
                   == '0x2e1a7d4d000000000000000000000000000000000000000000000006aaf7c8516d0c0000'  # function selector + encoded arg data



"""
test cases for frame rendering
"""

# lib
import time
import json
from urllib.parse import quote, unquote
import pytest
from flask import Flask

# src
from framelib import frame

app = Flask(__name__)


class TestRenderFrame(object):

    def test_one_button(self):
        with app.app_context():
            r = frame(button1='click')
            assert r.status_code == 200
            assert b'<meta property="fc:frame:button:1" content="click"/>' in r.data
            assert b'<meta property="fc:frame:button:2"' not in r.data

    def test_multiple_buttons(self):
        with app.app_context():
            r = frame(image='https://website.com/im.png', button1='click', button2='other', button3='back')
            assert r.status_code == 200
            assert b'<meta property="fc:frame:button:1" content="click"/>' in r.data
            assert b'<meta property="fc:frame:button:2" content="other"/>' in r.data
            assert b'<meta property="fc:frame:button:3" content="back"/>' in r.data
            assert b'<meta property="fc:frame:button:4"' not in r.data
            assert b'<meta property="fc:frame:image" content="https://website.com/im.png"/>' in r.data

    def test_aspect_ratio_default(self):
        with app.app_context():
            r = frame(image='https://website.com/im.png')
            assert r.status_code == 200
            assert b'<meta property="fc:frame:image" content="https://website.com/im.png"/>' in r.data
            assert b'<meta property="fc:frame:image:aspect_ratio"' not in r.data

    def test_aspect_ratio_1_1(self):
        with app.app_context():
            r = frame(image='https://website.com/im.png', aspect_ratio='1:1')
            assert r.status_code == 200
            assert b'<meta property="fc:frame:image" content="https://website.com/im.png"/>' in r.data
            assert b'<meta property="fc:frame:image:aspect_ratio" content="1:1"/>' in r.data

    def test_state(self):
        with app.app_context():
            s = quote(json.dumps({'app': 'state'}))
            r = frame(image='https://website.com/im.png', state=s)
            assert r.status_code == 200
            assert b'<meta property="fc:frame:image" content="https://website.com/im.png"/>' in r.data
            assert f'<meta property="fc:frame:state" content="{s}"/>'.encode('utf-8') in r.data

    def test_max_age_default(self):
        with app.app_context():
            r = frame(image='https://website.com/im.png')
            assert r.status_code == 200
            assert r.cache_control.max_age is None
            assert b'<meta property="fc:frame:image" content="https://website.com/im.png"/>' in r.data

    def test_max_age_short(self):
        with app.app_context():
            r = frame(image='https://website.com/im.png', max_age=60)
            assert r.status_code == 200
            assert r.cache_control.max_age == 60
            assert b'<meta property="fc:frame:image" content="https://website.com/im.png"/>' in r.data

    def test_max_age_zero(self):
        with app.app_context():
            r = frame(image='https://website.com/im.png', max_age=0)
            assert r.status_code == 200
            assert r.cache_control.max_age == 0
            assert b'<meta property="fc:frame:image" content="https://website.com/im.png"/>' in r.data



"""
test cases for frame message validation
"""

# lib
import pytest

# src
from framelib import validate_message
from framelib.neynar import get_frame_message
from framelib.hub import get_message


class TestValidateMessageNeynar(object):

    def test_button_1(self):
        # example signed message from fid 8268 clicking button 1
        msg = '0a4e080d10cc4018cbe1a230200182013f0a2068747470733a2f2f707974686f6e2d6672616d652e76657263656c2e6170702f10011a1908cc401214000000000000000000000000000000000000000112140101bf04a2e61cb24c9a66c047ac5ed175e1bed8180122403feee9d0c1392c1e5bc7bca49850f83735c53b4f60c88959ffc271123e333a196e963d15619125e6034acda36076c709182daa5625e4affe6df21866c204830828013220ad4520314a78bc4317c604a3324ebc25bd8215c3ac38342fd790b7905c291bd1'
        action = get_frame_message(msg, 'NEYNAR_API_DOCS')
        assert action.tapped_button.index == 1
        assert action.interactor.fid == 8268
        assert action.input is None


class TestValidateMessageHub(object):

    def test_button_1(self):
        # example signed message from fid 8268 clicking button 1
        msg = '0a4e080d10cc4018cbe1a230200182013f0a2068747470733a2f2f707974686f6e2d6672616d652e76657263656c2e6170702f10011a1908cc401214000000000000000000000000000000000000000112140101bf04a2e61cb24c9a66c047ac5ed175e1bed8180122403feee9d0c1392c1e5bc7bca49850f83735c53b4f60c88959ffc271123e333a196e963d15619125e6034acda36076c709182daa5625e4affe6df21866c204830828013220ad4520314a78bc4317c604a3324ebc25bd8215c3ac38342fd790b7905c291bd1'
        action = get_message(msg, 'https://nemes.farcaster.xyz:2281')  # public read only
        assert action.data.frameActionBody.buttonIndex == 1
        assert action.data.fid == 8268
        assert action.data.network == 'FARCASTER_NETWORK_MAINNET'
        assert action.data.frameActionBody.inputText == ''

    def test_button_1_neynar(self):
        # example signed message from fid 8268 clicking button 1
        msg = '0a4e080d10cc4018cbe1a230200182013f0a2068747470733a2f2f707974686f6e2d6672616d652e76657263656c2e6170702f10011a1908cc401214000000000000000000000000000000000000000112140101bf04a2e61cb24c9a66c047ac5ed175e1bed8180122403feee9d0c1392c1e5bc7bca49850f83735c53b4f60c88959ffc271123e333a196e963d15619125e6034acda36076c709182daa5625e4affe6df21866c204830828013220ad4520314a78bc4317c604a3324ebc25bd8215c3ac38342fd790b7905c291bd1'
        action = get_message(msg, 'https://hub-api.neynar.com', api_key='NEYNAR_API_DOCS')
        assert action.data.frameActionBody.buttonIndex == 1
        assert action.data.fid == 8268
        assert action.data.network == 'FARCASTER_NETWORK_MAINNET'
        assert action.data.frameActionBody.inputText == ''
